open Big_int
open Parser
open Pl

let eq = eq_big_int
let neq x y = not(eq x y)
let leq = le_big_int
let zero = zero_big_int
let one = unit_big_int
let read_big_int() = big_int_of_string(read_line())
let print_big_int x = print_string(string_of_big_int x)
let int = int_of_big_int

let rec lookup (f,i) = function
  | ((g,p),c)::_ when g = f && p i -> c
  | _ ::symbs                      -> lookup (f,i) symbs
  | []                             ->
    failwith("no symbol '"^Strng.to_string f^"' of arity "^string_of_int i)

let update env xs ys =
  Lst.zip2_with (fun x y -> ((x,(=)0),fun[] -> y)) xs ys @ env

let rec eval (env:((sym * (int -> bool)) * (big_int list -> big_int))list) = function
  | Const i         -> big_int_of_string(Strng.to_string i)
  | Ite(b,i,e)      ->
    if neq (eval env b) zero then eval env i
                             else eval env e
  | Call(f,es)      ->
    let call = lookup (f,Lst.length es) env in
    call (Lst.map (eval env) es)
  | Def(x,[],e1,e2) ->
    let call _ = eval env e1 in
    let env'   = ((x,(=)0),call) :: env in
    eval env' e2
  | Def(f,xs,e1,e2) ->
    let rec call vs = eval (update env' xs vs) e1
    and env' = ((f,(=)(Lst.length xs)),call) :: env in
    eval env' e2

let all _ = true

let penv = Lst.map (fun((f,a),call) -> ((Strng.of_string f,a),call)) [
  (("nl", (=)0),fun[]    -> print_newline();one);
  (("in", (=)0),fun[]    -> read_big_int());
  (("$",  (=)1),fun[x]   -> if Array.length Sys.argv < int x + 2
    then read_big_int()
    else big_int_of_string(Sys.argv.(int x+1)));
  (("not",(=)1),fun[x]   -> if neq x zero then zero else one);
  (("+",  (=)2),fun[x;y] -> add_big_int x y);
  (("-",  (=)2),fun[x;y] -> sub_big_int x y);
  (("*",  (=)2),fun[x;y] -> mult_big_int x y);
  (("div",(=)2),fun[x;y] -> div_big_int x y);
  (("mod",(=)2),fun[x;y] -> mod_big_int x y);
  (("eq", (=)2),fun[x;y] -> if eq x y then one else zero);
  (("leq",(=)2),fun[x;y] -> if leq x y then one else zero);
  (("@",   all),fun xs   -> one);
  (("out", all),fun xs   -> Lst.map print_big_int xs;one);
]

let run s =
  let input = s in
  match parse PlLexer.tokenize input with
    | None    -> failwith "lexing error"
    | Some ts ->
      let ts = Lst.filter (fun t -> not(PlLexer.is_comment t)) ts in
      match parse PlParser.ast ts with
        | None     -> failwith "parse error"
        | Some ast -> eval penv ast

let interpret file =
  let ss = File.read file in
  run ss
;;
if Array.length Sys.argv < 2 then (
  prerr_endline "usage: pl <file>";
  exit 1 
) else (
  interpret Sys.argv.(1);
)
